# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from .exceptions import ValidationTypeError
from .value_type_factory import ValueTypeFactory


class Bool(ValueTypeFactory):
    """ The Bool class encapsulates a Python bool. """

    def __init__(self, default=False, allow_none=False, getter=None, setter=None, **metadata):
        """ Initialise the object.

        :param default:
            is the default value of the bool.
        :param allow_none:
            ``True`` if ``None`` is a valid value.  The default is ``False``.
        :param getter:
            is the optional attribute getter.
        :param setter:
            is the optional attribute setter.
        :param \*\*metadata:
            is additional meta-data stored with the type.
        """

        super(Bool, self).__init__(default, allow_none, getter, setter,
                metadata)

    def validate(self, value):
        """ Validate a bool according to the constraints.  An exception is
        raised if the bool is invalid.

        :param value:
            the bool to validate.
        :return:
            the validated bool.
        """

        if value is None and self.allow_none:
            return value

        if not isinstance(value, bool):
            # See if it supports the bool protocol so that we avoid Python's
            # liberal bool convertions.
            if not hasattr(value, '__bool__'):
                raise ValidationTypeError(type(self), bool, value)

            value = bool(value)

        return value
