# Copyright (c) 2011 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from PyQt4.QtCore import QPoint, Qt
from PyQt4.QtGui import QWizard
from PyQt4.QtTest import QTest

from ....automate import AutomationError, IAutomatedDialog
from ....model import adapt, Adapter, notify_observers
from ....ui import IWizard

from ..widget_adapter import WidgetAdapter


@adapt(QWizard, to=IWizard)
class QWizardIWizardAdapter(WidgetAdapter):
    """ An adapter to implement IWizard for a QWizard. """

    def configure(self, properties):
        """ Configure the widget. """

        wizard = self.adaptee

        wizard.currentIdChanged.connect(self._current_id_changed)
        wizard.button(QWizard.FinishButton).clicked.connect(self._on_finished)

        super(QWizardIWizardAdapter, self).configure(properties)

    def execute(self):
        """ Execute the wizard.

        :return:
            ``True`` if the user accepted the wizard.
        """

        return bool(self.adaptee.exec_())

    @IWizard.current_page.getter
    def current_page(self):
        """ Invoked to get the current page. """

        return self.adaptee.currentPage()

    @current_page.setter
    def current_page(self, value):
        """ Invoked to set the current page. """

        wizard = self.adaptee

        for id in wizard.pageIds():
            if wizard.page(id) is value:
                wizard.setStartId(id)
                wizard.restart()
                break

    @IWizard.views.getter
    def views(self):
        """ Invoked to get the views (i.e. pages). """

        wizard = self.adaptee

        return tuple(wizard.page(page_id) for page_id in wizard.pageIds())

    @views.setter
    def views(self, value):
        """ Invoked to set the views (i.e. pages). """

        wizard = self.adaptee

        # Remove any existing pages.
        for page_id in wizard.pageIds():
            wizard.removePage(page_id)

        # Add the new pages.
        first_page = None
        final_page = wizard.options() & QWizard.HaveFinishButtonOnEarlyPages

        for page in value:
            if first_page is None:
                first_page = page

            if final_page:
                page.setFinalPage(True)

            wizard.addPage(page)

        self._qt4_old_page = None
        self.current_page = first_page

    def _current_id_changed(self, id):
        """ Invoked then the current page changes. """

        old_page = self._qt4_old_page
        new_page = self._qt4_old_page = self.adaptee.page(id)

        notify_observers('current_page', self, new_page, old_page)

    def _on_finished(self):
        """ Invoked when the Finish button is clicked. """

        self.finished = True


@adapt(QWizard, to=IAutomatedDialog)
class QWizardIAutomatedDialogAdapter(Adapter):
    """ An adapter to implement IAutomatedDialog for the Qt dialog. """

    def simulate_click(self, id, delay, button):
        """ Simulate the user clicking a button.

        :param id:
            is the (possibly scoped) identifier of the widget.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param button:
            is the :class:`~PyQt4.QtGui.QWizard.WizardButton` to click.
        """

        # Find the button widget.
        # FIXME: Use an abstract button name instead.
        button_widget = self.adaptee.button(button)

        if button_widget is None:
            raise AutomationError(id, 'click',
                    "no {0} button in wizard".format(hex(button)))

        QTest.mouseClick(button_widget, Qt.LeftButton, Qt.NoModifier, QPoint(),
                delay)
