# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Callable, get_attribute_type, MappingProxy, Model
from ..toolkits import Toolkit

from .i_controller import IController
from .editor_factory import EditorFactory
from .view_factory import ViewFactory


class Bindings(Model):
    """ The Bindings class represents set of bindings between widgets of a user
    interface and attributes of a model.
    """

    # The controller factory.  This must return an object that implements or
    # can be adapted to the :class:`~dip.ui.IController` interface.
    controller_factory = Callable()

    def __init__(self, **bindings):
        """ Initialise the bindings.

        :param \*\*bindings:
            the bindings specified as keyword arguments.  The argument name is
            the object name of the widget that the attribute will be bound to.
            The argument value is the binding.  If the argument value is a
            string then an :class:`~dip.ui.EditorFactory` instance will be
            automatically created for it.
        """

        self._bindings = bindings

    def bind(self, view, model, controller=None):
        """ Bind a view to a model and optional controller.

        :param view:
            is the view.
        :param model:
            is the model which is either a :class:`~dip.model.Model` instance
            or a Python mapping object.
        :param controller:
            is the optional controller.  If this is ``None`` then
            :attr:`~dip.ui.Bindings.controller_factory` will be called to
            create one.
        """

        # Make sure we have a Model for the model.
        if not isinstance(model, Model):
            model = MappingProxy(model)

        # Make sure we have a controller.
        if controller is None:
            # Pass the model rather than set it explicitly to allow the
            # controller to use observe() as a decorator.
            controller = self.controller_factory(model=model)

        IController(controller).view = view

        finalize = []

        for view_name, itm in self._bindings.items():
            # Find the sub-view in the user interface.
            subview = Toolkit.find_view(view, view_name)

            # Automatically create a factory if needed.
            if isinstance(itm, EditorFactory):
                name = itm.bind_to
                subview_factory = itm
            elif isinstance(itm, str):
                name = itm

                # Create an editor factory based on the widget type.
                editor_factory_factory = Toolkit.declarative_factory_for_view(
                        subview)
                subview_factory = editor_factory_factory(itm, id=view_name)
            elif isinstance(itm, ViewFactory):
                name = None
                subview_factory = itm
            else:
                raise TypeError(
                        "the binding for '{0}' must be either an instance of "
                        "a ViewFactory sub-class or a string, not "
                        "'{1}'".format(view_name, type(itm).__name__))

            # Make sure the widget can be adapted to the right interface.
            iview = subview_factory.view_interface(subview)

            # Update the factory with the bound widget's identifier so that it
            # doesn't get changed when the widget is configured.
            subview_factory.id = iview.id

            if name is None:
                subview_factory.bind_view(subview, controller)
            else:
                subview_factory.bind_editor(subview, model, name, controller)

            finalize.append((subview_factory, subview))

        for subview_factory, subview in finalize:
            subview_factory.finalize(subview)

    @controller_factory.default
    def controller_factory(self):
        """ Invoked to return the default controller factory. """

        from .controller import Controller

        return Controller
