# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import (Any, Callable, get_model_types, Instance, List, MetaModel,
        Str)
from ..toolkits import Toolkit

from .i_editor_binding import IEditorBinding
from .editor_factory import EditorFactory
from .list_column import ListColumn
from .i_list_editor import IListEditor


class ListEditor(EditorFactory):
    """ The ListEditor class implements a list editor factory for list
    attributes.  If an item of the list is a simple type then there will be a
    a single column.  If an item of the list is a model then there may be a
    column for each attribute of the model.
    """

    # The list of columns.  If a string is given then it is assumed to be the
    # name of an attribute in the row model.
    columns = List(Instance(ListColumn, str))

    # The factory for the view created to edit a value in any column that does
    # not specify its own editor factory.  It will default to the factory
    # provided by the toolkit for the column type.
    editor_factory = Callable()

    # When called this will create the data for a new item in the list.  If a
    # factory isn't provided then a value based on the default value of the
    # type of the list's elements is used.
    item_factory = Callable()

    # The specific interface implemented by the view.
    view_interface = IListEditor

    def configure(self, view):
        """ Reimplemented to configure the IListEditor specific attributes.

        :param view:
            is the view to configure.
        """

        ilistedit = IListEditor(view)

        # Create complete column specifications.
        columns = []

        # See what we have a list of.
        attribute_type = IEditorBinding(view).attribute_type
        element_type = attribute_type.element_type
        item_factory = self.item_factory

        nr_columns = len(self.columns)
        redundant_column_type = False

        if isinstance(element_type, MetaModel):
            model_types = get_model_types(element_type)

            if nr_columns == 0:
                # Create a column for each attribute of the model.
                columns = [ListColumn(name, column_type=atype,
                                label=self.get_natural_name(name))
                        for name, atype in model_types.items()]
            else:
                # Convert strings to columns.
                for col in self.columns:
                    if isinstance(col, str):
                        column = ListColumn(col)
                    elif col.column_type is None:
                        column = ListColumn(col.bind_to,
                                editor_factory=col.editor_factory,
                                label=col.label)
                    else:
                        redundant_column_type = True

                    try:
                        column.column_type = model_types[column.bind_to]
                    except KeyError:
                        raise AttributeError(
                                "'{0}' has no attribute '{1}'".format(
                                        element_type.__name__, column.bind_to))

                    # Make sure there is a label.
                    if column.label is None:
                        column.label = self.get_natural_name(column.bind_to)

                    columns.append(column)

            # Make sure there is an item factory.
            if item_factory is None:
                item_factory = element_type

        else:
            if nr_columns == 0:
                column = ListColumn(column_type=element_type)
            elif nr_columns == 1:
                col = self.columns[0]

                if col.column_type is None:
                    column = ListColumn(column_type=element_type,
                            editor_factory=col.editor_factory,
                            label=col.label)
                else:
                    redundant_column_type = True
            else:
                raise AttributeError(
                        "too many column definitions for a single column list")

            # Make sure there is a label.
            if column.label is None:
                column.label = self.get_natural_name(attribute_type.name)

            columns.append(column)

        if redundant_column_type:
            raise AttributeError(
                    "'column_type' should not be specified because it can be "
                    "inferred")

        if len(columns) == 0:
            raise AttributeError(
                    "unable to infer how many columns the list editor should "
                    "have")

        # Make sure each column has a type and an appropriate editor factory.
        for col in columns:
            if col.column_type is None or isinstance(col.column_type, Any):
                col.column_type = Str()

            if col.editor_factory is None:
                if self.editor_factory is None:
                    col.editor_factory = Toolkit.editor_factory_for_attribute_type(col.column_type)
                else:
                    col.editor_factory = self.editor_factory

        # Make sure that there is an item factory.  If we haven't got one yet
        # it's because an item is a simple value (i.e. not a model).
        if item_factory is None:
            item_factory = columns[0].column_type.get_default_value

        ilistedit.columns = columns
        ilistedit.item_factory = item_factory

        super(ListEditor, self).configure(view)
