# Copyright (c) 2011 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import Instance, List, Model, observe, Str, Trigger


class TestObserver(unittest.TestCase):
    """ This tests the observer function. """

    def log(self, change):
        """ Remember the change passed for later checking by check_log(). """

        self._logged_change = change
        self._logged_count += 1

    def check_log(self, count, model, name, new, old):
        """ Check the logged change equals the given values. """

        self.assertEqual(self._logged_count, count)

        change = self._logged_change
        if change is not None:
            self.assertIs(change.model, model)
            self.assertEqual(change.new, new)
            self.assertEqual(change.old, old)

    def setUp(self):
        """ Initialise the log. """

        self._logged_change = None
        self._logged_count = 0

    def test_observer(self):
        """ Observing a model attribute. """

        class Klass(Model):
            attr = Str()

        k = Klass()

        observe('attr', k, self.log)

        k.attr = 'observed'
        self.check_log(1, k, 'attr', 'observed', '')

    def test_observer_trigger(self):
        """ Observing a trigger attribute. """

        class Klass(Model):
            attr = Trigger()

        k = Klass()

        observe('attr', k, self.log)

        k.attr = 'observed'
        self.check_log(1, k, 'attr', 'observed', None)

    def test_observer_instances(self):
        """ Observing a model attribute with multiple instances. """

        class Klass(Model):
            attr = Str()

        k1 = Klass()
        k2 = Klass()

        observe('attr', k1, self.log)

        k2.attr = 'ignored'
        self.check_log(0, None, 'attr', '', '')

    def test_observer_remove(self):
        """ Removing an observer. """

        class Klass(Model):
            attr = Str()

        k = Klass()

        observe('attr', k, self.log)

        k.attr = 'observed'
        self.check_log(1, k, 'attr', 'observed', '')

        observe('attr', k, self.log, remove=True)

        k.attr = 'unobserved'
        self.check_log(1, k, 'attr', 'observed', '')

    def test_observer_unchanged_fundamental_type(self):
        """ Observing an unchanged fundamental attribute. """

        class Klass(Model):
            attr = Str()

        k = Klass()

        observe('attr', k, self.log)

        k.attr = 'observed'
        self.check_log(1, k, 'attr', 'observed', '')

        k.attr = k.attr
        self.check_log(1, k, 'attr', 'observed', '')

    def test_observer_unchanged_collection(self):
        """ Observing an unchanged collection attribute. """

        class Data(Model):
            datum = Str()

            # This will cause the test to fail if it is wrongly using != rather
            # than 'is not'.
            def __ne__(self, other):
                return True

        data = Data()

        class Klass(Model):
            attr = Instance(Data)

        k = Klass()

        observe('attr', k, self.log)

        k.attr = data
        self.check_log(1, k, 'attr', data, None)

        k.attr = k.attr
        self.check_log(1, k, 'attr', data, None)

    def test_observer_mutable(self):
        """ Observing a model mutable attribute. """

        class Klass(Model):
            attr = List()

        k = Klass()

        observe('attr', k, self.log)

        k.attr = ['assigned']
        self.check_log(1, k, 'attr', ['assigned'], [])

    def test_decorator(self):
        """ Observing using the decorator. """

        test_case = self

        class SubKlass(Model):
            attr = Str()

        class Klass(Model):
            sub = Instance(SubKlass)

            @sub.default
            def sub(self):
                return SubKlass()

            @observe('sub.attr')
            def _sub_attr_changed(self, change):
                test_case.log(change)

        k = Klass()
        k.sub.attr = 'observed'
        self.check_log(1, k.sub, 'attr', 'observed', '')

    def test_multi_decorator(self):
        """ Observing using the decorator more than once. """

        test_case = self

        class SubKlass(Model):
            attr = Str()

        class Klass(Model):
            sub = Instance(SubKlass)

            @sub.default
            def sub(self):
                return SubKlass()

            @observe('sub.attr')
            @observe('sub.attr')
            def _sub_attr_changed(self, change):
                test_case.log(change)

        k = Klass()
        k.sub.attr = 'observed'
        self.check_log(2, k.sub, 'attr', 'observed', '')

    def test_wildcard_decorator(self):
        """ Observing using the decorator with a wildcard name. """

        test_case = self

        class SubKlass(Model):
            attr1 = Str()

            attr2 = Str()

        class Klass(Model):
            sub = Instance(SubKlass)

            @sub.default
            def sub(self):
                return SubKlass()

            @observe('sub.*')
            def _sub_attr_changed(self, change):
                test_case.log(change)

        k = Klass()
        k.sub.attr1 = 'observed1'
        self.check_log(1, k.sub, 'attr1', 'observed1', '')
        k.sub.attr2 = 'observed2'
        self.check_log(2, k.sub, 'attr2', 'observed2', '')

    def test_common_type(self):
        """ Observing attributes with the same type object. """

        test_case = self

        attr_type = Str()

        class Klass(Model):
            attr1 = attr_type

            @observe('attr1')
            def _attr1_changed(self, change):
                test_case.log(change)

            attr2 = attr_type

            @observe('attr2')
            def _attr2_changed(self, change):
                test_case.log(change)

        k = Klass()
        k.attr1 = 'observed1'
        self.check_log(1, k, 'attr1', 'observed1', '')
        k.attr2 = 'observed2'
        self.check_log(2, k, 'attr2', 'observed2', '')


if __name__ == '__main__':
    unittest.main()
