# Copyright (c) 2010 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import Model, observe, Str, Tuple, ValidationError


class Klass(Model):
    """ A class for test models. """

    default_tuple = Tuple()

    explicit_default_tuple = Tuple(default=(None, ))

    attr = Tuple(Str())

    def __init__(self):
        super(Klass, self).__init__()

        self.change = None
        self.changes = 0
        observe('attr', self, self._observed)

    def _observed(self, change):
        self.change = change
        self.changes += 1


class TestTupleType(unittest.TestCase):
    """ This tests the Tuple type. """

    def check_observed(self, model, new, old):
        """ Check that a test model observed a change. """

        self.assertEquals(model.changes, 1)

        change = model.change

        self.assertIs(change.model, model)
        self.assertEquals(change.name, 'attr')
        self.assertEquals(change.new, new)
        self.assertEquals(change.old, old)

    def check_unobserved(self, model):
        """ Check that a test model hasn't observed a change. """

        self.assertEquals(model.changes, 0)

    def test_invalid_tuple_type(self):
        """ Invalid Tuple type. """

        with self.assertRaises(TypeError):
            Tuple(0)

    def test_default_tuple_type_explicit_default(self):
        """ Default Tuple explicit default value. """

        k = Klass()
        self.assertTupleEqual(k.explicit_default_tuple, (None, ))

    def test_tuple_subclass_default(self):
        """ Default Tuple sub-class. """

        class SubKlass(Klass):

            default_tuple = (1, )

        k = SubKlass()
        self.assertTupleEqual(k.default_tuple, (1, ))

    def test_default_tuple_type_default_value(self):
        """ Default Tuple type default value. """

        k = Klass()
        self.assertTupleEqual(k.default_tuple, ())

    def test_default_tuple_type_valid_value(self):
        """ Default Tuple type valid value. """

        k = Klass()
        k.default_tuple = (1, True, 'one')
        self.assertTupleEqual(k.default_tuple, (1, True, 'one'))

    def test_default_tuple_type_invalid_value(self):
        """ Default Tuple type invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.default_tuple = None

    def test_tuple_type_valid_default_value(self):
        """ Tuple type valid default value. """

        k = Klass(attr=('abc', 'def'))
        self.assertTupleEqual(k.attr, ('abc', 'def'))

    def test_tuple_type_invalid_default_value(self):
        """ Tuple type invalid default value. """

        with self.assertRaises(ValidationError):
            Klass(attr=(False, ))

    def test_tuple_type_rebind_valid_value(self):
        """ Tuple type rebind valid value. """

        k = Klass()
        k.attr = ('abc', 'def')
        self.assertTupleEqual(k.attr, ('abc', 'def'))
        self.check_observed(k, ('abc', 'def'), ())

    def test_tuple_type_rebind_unchanged_value(self):
        """ Tuple type rebind unchanged value. """

        k = Klass()
        k.attr = ()
        self.assertTupleEqual(k.attr, ())
        self.check_unobserved(k)

    def test_tuple_type_rebind_invalid_value(self):
        """ Tuple type rebind invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = (False, )

    def test_tuple_type_rebind_invalid_tuple(self):
        """ Tuple type rebind invalid tuple. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = k.default_tuple


if __name__ == '__main__':
    unittest.main()
